# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from __future__ import absolute_import, print_function

import json

from invenio_rest import InvenioREST
from invenio_rest.decorators import require_content_types


def test_require_content_types(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route("/", methods=["POST"])
    @require_content_types("application/json", "text/plain")
    def test_view():
        return "OK"

    with app.test_client() as client:
        res = client.post("/", content_type="application/json", data="{}")
        assert res.status_code == 200
        res = client.post("/", content_type="text/plain", data="test")
        assert res.status_code == 200
        res = client.post("/", content_type="application/json;charset=utf-8", data="{}")
        assert res.status_code == 200
        res = client.post("/", content_type="application/xml", data="<d></d>")
        assert res.status_code == 415
        data = json.loads(res.get_data(as_text=True))
        assert data["status"] == 415
        assert "application/json" in data["message"]
        assert "text/plain" in data["message"]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2025 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import json
from datetime import datetime
from importlib.metadata import distribution
from urllib.parse import urlencode

import pytest
from flask import Flask, abort, make_response, request
from flask.json import jsonify
from werkzeug.http import quote_etag, unquote_etag

from invenio_rest import ContentNegotiatedMethodView, InvenioREST
from invenio_rest.errors import RESTException


def test_version():
    """Test version import."""
    from invenio_rest import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioREST(app)
    assert "invenio-rest" in app.extensions

    app = Flask("testapp")
    ext = InvenioREST()
    assert "invenio-rest" not in app.extensions
    ext.init_app(app)
    assert "invenio-rest" in app.extensions


def test_error_handlers(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route(
        "/<int:status_code>",
        methods=["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "TRACE", "OPTIONS"],
    )
    def test_view(status_code):
        abort(status_code)

    error_codes = [
        400,
        401,
        403,
        404,
        405,
        406,
        409,
        410,
        412,
        422,
        415,
        429,
        500,
        501,
        502,
        503,
        504,
    ]

    with app.test_client() as client:
        verbs = [
            client.get,
            client.post,
            client.put,
            client.delete,
            client.patch,
            client.trace,
            client.options,
            client.get,
        ]
        for s in error_codes:
            for verb in verbs:
                res = verb("/{0}".format(s))
                assert res.status_code == s
                if distribution("werkzeug").version == "0.14.1" and s == 412:
                    # skip test because of the following issue:
                    # https://github.com/pallets/werkzeug/issues/1231
                    continue
                if verb != client.head:
                    data = json.loads(res.get_data(as_text=True))
                    assert data["status"] == s
                    assert data["message"]


def test_custom_httpexception(app):
    """Test custom RESTException."""
    InvenioREST(app)

    class CustomBadRequest(RESTException):
        code = 400
        description = "test"

    @app.route("/")
    def test_view():
        raise CustomBadRequest()

    with app.test_client() as client:
        res = client.get("/")
        assert res.status_code == 400
        data = json.loads(res.get_data(as_text=True))
        assert data["message"] == "test"
        assert data["status"] == 400


def test_cors(app):
    """Test CORS support."""
    app.config["REST_ENABLE_CORS"] = True
    app.config["CORS_ALWAYS_SEND"] = False

    InvenioREST(app)

    @app.route("/")
    def cors_test():
        return "test"

    with app.test_client() as client:
        res = client.get("/")
        assert not res.headers.get("Access-Control-Allow-Origin", False)
        assert not res.headers.get("Access-Control-Expose-Headers", False)
        res = client.get("/", headers=[("Origin", "http://example.com")])
        assert res.headers["Access-Control-Allow-Origin"] == "*"
        assert len(res.headers["Access-Control-Expose-Headers"].split(",")) == len(
            app.config["CORS_EXPOSE_HEADERS"]
        )


def _obj_to_json_serializer(data, code=200, headers=None):
    if data:
        res = jsonify(data)
        res.cache_control.no_cache = True
        res.set_etag("abc")
    else:
        res = make_response()
    res.status_code = code
    return res


class _ObjectItem(ContentNegotiatedMethodView):
    view_name = "object_item"

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(_ObjectItem, self).__init__(*args, **kwargs)
        self.modified = datetime.utcnow()

    def get(self, id, **kwargs):
        """GET a resource."""
        self.check_etag("abc", weak="weak_etags" in request.args)
        self.check_if_modified_since(self.modified)
        if id == 42:
            abort(404)
        # return {'id': id, 'method': 'GET'}, 200
        res = self.make_response({"id": id, "method": "GET"}, 200)
        res.last_modified = self.modified
        return res

    def put(self, id, **kwargs):
        """PUT a resource."""
        self.check_etag("abc", weak="weak_etags" in request.args)
        if id == 42:
            return self.make_response(None, code=404)
        return self.make_response({"id": id, "method": "PUT"}, 200)

    def patch(self, id, **kwargs):
        """PATCH a resource."""
        self.check_etag("abc", weak="weak_etags" in request.args)
        if id == 42:
            return None, 404
        return self.make_response({"id": id, "method": "PATCH"})

    def post(self, id, **kwargs):
        """POST a resource."""
        self.check_etag("abc", weak="weak_etags" in request.args)
        if id == 42:
            abort(404)
        return {"id": id, "method": "POST"}

    def delete(self, id, **kwargs):
        """DELETE a resource."""
        self.check_etag("abc", weak="weak_etags" in request.args)
        if id == 42:
            abort(404)
        return {"id": id, "method": "DELETE"}


def test_content_negotiation_method_view_global_serializers(app):
    """Test ContentNegotiationMethodView. Test default serializers."""
    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "serializers": {
                "application/json": _obj_to_json_serializer,
            }
        },
    )


def test_content_negotiation_method_view_method_serializers(app):
    """Test method serializers with no global serializers."""
    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "method_serializers": {
                "GET": {
                    "application/json": _obj_to_json_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                },
                "POST": {
                    "application/json": _obj_to_json_serializer,
                },
                "PATCH": {
                    "application/json": _obj_to_json_serializer,
                },
                "DELETE": {
                    "application/json": _obj_to_json_serializer,
                },
            }
        },
    )


def test_content_negotiation_method_view_method_serializers_2(app):
    """Check that method serializers override global ones."""

    def failing_serializer(data, code=200, headers=None):
        raise Exception("Method serializer was not called.")

    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "serializers": {
                "application/json": failing_serializer,
            },
            "method_serializers": {
                "GET": {
                    "application/json": _obj_to_json_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                },
                "POST": {
                    "application/json": _obj_to_json_serializer,
                },
                "PATCH": {
                    "application/json": _obj_to_json_serializer,
                },
                "DELETE": {
                    "application/json": _obj_to_json_serializer,
                },
            },
        },
    )


def test_content_negotiation_method_view_mixed_serializers(app):
    """Test ContentNegotiationMethodView. Test partial method serializers"""
    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "serializers": {
                "application/json": _obj_to_json_serializer,
            },
            "method_serializers": {
                "GET": {
                    "application/json": _obj_to_json_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                },
            },
        },
    )


def test_default_media_type(app):
    """Test setting default media type."""

    def failing_serializer(data, code=200, headers=None):
        raise Exception("Wrong serializer was called.")

    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "serializers": {
                "application/mpeg": failing_serializer,
                "application/json": _obj_to_json_serializer,
            },
            "default_media_type": "application/json",
        },
    )


def test_default_media_type_for_all_methods(app):
    """Test setting default serializer for all methods."""

    def failing_serializer(data, code=200, headers=None):
        raise Exception("Wrong method serializer was called.")

    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "method_serializers": {
                "GET": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "POST": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "PATCH": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "DELETE": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
            },
            "default_media_type": "application/json",
        },
    )


def test_default_method_media_type(app):
    """Test setting default media type for each method."""

    def failing_serializer(data, code=200, headers=None):
        raise Exception("Wrong method serializer was called.")

    _subtest_content_negotiation_method_view(
        app,
        _ObjectItem,
        params={
            "method_serializers": {
                "GET": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "POST": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "PATCH": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
                "DELETE": {
                    "application/json": _obj_to_json_serializer,
                    "application/mpeg": failing_serializer,
                },
            },
            "default_method_media_type": {
                "GET": "application/json",
                "PUT": "application/json",
                "POST": "application/json",
                "PATCH": "application/json",
                "DELETE": "application/json",
            },
            "default_media_type": "application/doesnotexist",
        },
    )


def test_default_method_media_type_error(app):
    """Test that there is an error when multiple serializers are given with no
    default media type.
    """
    with pytest.raises(ValueError):
        _ObjectItem(
            method_serializers={
                "GET": {
                    "application/json": _obj_to_json_serializer,
                    "application/json-patch+json": _obj_to_json_serializer,
                },
                "PUT": {
                    "application/json": _obj_to_json_serializer,
                    "application/json-patch+json": _obj_to_json_serializer,
                },
                "POST": {
                    "application/json": _obj_to_json_serializer,
                    "application/json-patch+json": _obj_to_json_serializer,
                },
                "PATCH": {
                    "application/json": _obj_to_json_serializer,
                    "application/json-patch+json": _obj_to_json_serializer,
                },
                "DELETE": {
                    "application/json": _obj_to_json_serializer,
                    "application/json-patch+json": _obj_to_json_serializer,
                },
            },
        )


def test_default_media_type_error(app):
    """Test that there is an error when multiple serializers are given with no
    default media type.
    """
    with pytest.raises(ValueError):
        _ObjectItem(
            serializers={
                "application/json": _obj_to_json_serializer,
                "application/json-patch+json": _obj_to_json_serializer,
            }
        )


def test_get_method_serializers(app):
    """Test get request method serializers."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            DELETE={
                "*/*": "any-delete",
            },
            GET={"text/plain": "plain-get", "text/html": "html-get"},
            POST={"text/plain": "plain-post", "text/html": "html-post"},
        ),
        serializers={
            "text/plain": "plain",
            "text/html": "html",
        },
        default_media_type="application/json",
        default_method_media_type=dict(
            GET="text/plain",
            POST="text/html",
        ),
    )
    # Default serializers
    assert v.get_method_serializers("GET")[0] == v.method_serializers["GET"]
    assert v.get_method_serializers("HEAD")[0] == v.method_serializers["GET"]
    assert v.get_method_serializers("POST")[0] == v.method_serializers["POST"]
    assert v.get_method_serializers("PUT")[0] == v.serializers

    # Default media type
    assert v.get_method_serializers("GET")[1] == v.default_method_media_type["GET"]
    assert v.get_method_serializers("HEAD")[1] == v.default_method_media_type["GET"]
    assert v.get_method_serializers("POST")[1] == v.default_method_media_type["POST"]
    assert v.get_method_serializers("PUT")[1] == v.default_media_type

    assert v.get_method_serializers("DELETE")[0] == v.method_serializers["DELETE"]


def test_match_serializers_headers(app):
    """Test match serializers headers."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            DELETE={
                "*/*": "any-delete",
            },
            GET={"application/json": "json-get", "application/marcxml+xml": "xml-get"},
            POST={
                "application/json": "json-post",
                "application/marcxml+xml": "xml-post",
            },
        ),
        serializers={
            "application/json": "json",
            "application/marcxml+xml": "xml",
        },
        default_media_type="application/json",
        default_method_media_type=dict(
            GET="application/json",
            POST="application/marcxml+xml",
        ),
    )

    tests = [
        # Should choose application/json
        ("text/plain,application/json,*/*", "GET", "json-get"),
        ("text/plain,application/json,*/*", "HEAD", "json-get"),
        ("text/plain,application/json,*/*", "POST", "json-post"),
        ("text/plain,application/json,*/*", "PUT", "json"),
        # Should choose application/json (even with lower quality)
        ("text/plain,application/json; q=0.5, */*", "GET", "json-get"),
        ("text/plain,application/json; q=0.5, */*", "HEAD", "json-get"),
        ("text/plain,application/json; q=0.5, */*", "POST", "json-post"),
        ("text/plain,application/json; q=0.5, */*", "PUT", "json"),
        # Should choose application/marcxml+xml
        ("text/plain,application/marcxml+xml,*/*", "GET", "xml-get"),
        ("text/plain,application/marcxml+xml,*/*", "HEAD", "xml-get"),
        ("text/plain,application/marcxml+xml,*/*", "POST", "xml-post"),
        ("text/plain,application/marcxml+xml,*/*", "PUT", "xml"),
        # Should choose default defined by default_media_type and
        # default_method_media_type
        ("text/plain,*/*", "GET", "json-get"),
        ("text/plain,*/*", "HEAD", "json-get"),
        ("text/plain,*/*", "POST", "xml-post"),
        ("text/plain,*/*", "PUT", "json"),
        # Should choose none
        ("text/plain", "GET", None),
        ("text/plain", "HEAD", None),
        ("text/plain", "POST", None),
        ("text/plain", "PUT", None),
        # Should choose default
        (None, "GET", "json-get"),
        (None, "HEAD", "json-get"),
        (None, "POST", "xml-post"),
        (None, "PUT", "json"),
        # Should choose application/json
        ("application/json", "GET", "json-get"),
        # Should choose highest quality
        ("application/json; q=0.5, application/marcxml+xml", "GET", "xml-get"),
        ("application/json, application/marcxml+xml; q=0.5", "GET", "json-get"),
        ("application/json; q=0.4, application/marcxml+xml; q=0.6", "GET", "xml-get"),
        ("application/marcxml+xml; q=0.6, application/json; q=0.4", "GET", "xml-get"),
        # Test delete
        (
            "application/marcxml+xml; q=0.6, application/json; q=0.4",
            "DELETE",
            "any-delete",
        ),
        ("video/mp4", "DELETE", "any-delete"),
    ]

    for accept, method, expected_serializer in tests:
        params = dict(headers=[("Accept", accept)]) if accept else {}
        _test_march_serializers(app, v, params, method, accept, expected_serializer)


def test_match_serializers_query_arg(app):
    """Test match serializers query argument."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            GET={"application/json": "json-get", "application/marcxml+xml": "xml-get"}
        ),
        serializers_query_aliases={
            "json": "application/json",
            "marcxml": "application/marcxml+xml",
        },
        default_media_type="application/marcxml+xml",
        default_method_media_type=dict(
            GET="application/marcxml+xml",
        ),
    )

    # enable serialization by query arg
    app.config["REST_MIMETYPE_QUERY_ARG_NAME"] = "format"
    InvenioREST(app)

    arg_name = app.config["REST_MIMETYPE_QUERY_ARG_NAME"]
    tests = [
        # Should serialize to marcxml
        (urlencode({arg_name: "marcxml"}), "application/json", "xml-get"),
        # Should serialize to json
        (urlencode({arg_name: "json"}), "application/marcxml+xml", "json-get"),
        # Should serialize to json
        (urlencode({}), "application/json", "json-get"),
        # Should serialize to default (marcxml)
        (urlencode({"wrong_name": "marcxml"}), None, "xml-get"),
    ]

    for arg, accept, expected_serializer in tests:
        params = dict(headers=[("Accept", accept)], query_string=arg) if accept else {}
        _test_march_serializers(app, v, params, "GET", accept, expected_serializer)

    # disable query arg, should serialize by headers even if query arg passed
    app.config["REST_MIMETYPE_QUERY_ARG_NAME"] = None
    InvenioREST(app)

    accept = "application/json"
    params = dict(headers=[("Accept", accept)], query_string="format=marcxml")
    _test_march_serializers(app, v, params, "GET", accept, "json-get")


def _test_march_serializers(
    app,
    view_method,
    request_params,
    request_method,
    request_accept_header,
    expected_serializer,
):
    with app.test_request_context(**request_params):
        assert (
            view_method.match_serializers(
                *view_method.get_method_serializers(request_method)
            )
            == expected_serializer
        ), "Accept: {0}, Method: {1}".format(request_accept_header, request_method)


def _subtest_content_negotiation_method_view(app, content_negotiated_class, params):
    """Test a given ContentNegotiatedMethodView subclass."""
    app.add_url_rule(
        "/objects/<int:id>",
        view_func=content_negotiated_class.as_view(
            content_negotiated_class.view_name, **params
        ),
    )

    with app.test_client() as client:
        read_methods = [client.get, client.head]
        write_methods = [client.patch, client.put, client.post, client.delete]
        all_methods = read_methods + write_methods
        method_names = {
            "GET": client.get,
            "POST": client.post,
            "PUT": client.put,
            "PATCH": client.patch,
            "DELETE": client.delete,
        }

        def check_normal_response(res, method):
            if method != client.head:
                parsed = json.loads(res.get_data(as_text=True))
                expected = {"id": 1, "method": parsed["method"]}
                assert parsed == expected
                # check that the right method was called
                assert method_names[parsed["method"]] == method
                assert res.content_type == "application/json"
            assert res.status_code == 200
            # check that Cache-Control header is set
            assert res.cache_control.no_cache
            # check that the ETag is correct
            assert unquote_etag(res.headers["ETag"]) == unquote_etag(quote_etag("abc"))

        def check_304_response(res):
            assert res.status_code == 304
            # check that Cache-Control header is set
            assert res.cache_control.no_cache
            # check that the ETag is correct
            assert unquote_etag(res.headers["ETag"]) == unquote_etag(quote_etag("abc"))

        # check valid call with condition
        headers = [("Accept", "application/json")]
        for method in all_methods:
            res = method("/objects/1", headers=headers)
            check_normal_response(res, method)

        # check valid call without condition
        for method in all_methods:
            res = method("/objects/1")
            check_normal_response(res, method)

        # check that non accepted mime types are not accepted
        headers = [("Accept", "application/xml")]
        for method in all_methods:
            res = method("/objects/1", headers=headers)
            assert res.status_code == 406

        # check that errors are forwarded properly
        headers = [("Accept", "application/json")]
        for method in all_methods:
            res = method("/objects/42", headers=headers)
            assert res.status_code == 404

        # check Matching If-None-Match
        headers_nonmatch_match = [
            ("Accept", "application/json"),
            ("If-None-Match", '"xyz", "abc"'),
        ]
        headers_nonmatch_star = [
            ("Accept", "application/json"),
            ("If-None-Match", '"xyz", "*"'),
        ]
        for method in read_methods:
            res = method("/objects/1", headers=headers_nonmatch_match)
            check_304_response(res)
            res = method("/objects/1", headers=headers_nonmatch_star)
            check_304_response(res)

        for method in write_methods:
            res = method("/objects/1", headers=headers_nonmatch_match)
            assert res.status_code == 412
            res = method("/objects/1", headers=headers_nonmatch_star)
            assert res.status_code == 412

        # check Matching If-None-Match with weak ETags
        headers_nonmatch_match = [
            ("Accept", "application/json"),
            ("If-None-Match", 'W/"xyz", W/"abc"'),
        ]
        headers_nonmatch_star = [
            ("Accept", "application/json"),
            ("If-None-Match", 'W/"xyz", "*"'),
        ]
        query_string_weak_etags = {"weak_etags": True}
        for method in read_methods:
            res = method(
                "/objects/1",
                headers=headers_nonmatch_match,
                query_string=query_string_weak_etags,
            )
            check_304_response(res)
            res = method(
                "/objects/1",
                headers=headers_nonmatch_star,
                query_string=query_string_weak_etags,
            )
            check_304_response(res)

        for method in write_methods:
            res = method(
                "/objects/1",
                headers=headers_nonmatch_match,
                query_string=query_string_weak_etags,
            )
            assert res.status_code == 412
            res = method(
                "/objects/1",
                headers=headers_nonmatch_star,
                query_string=query_string_weak_etags,
            )
            assert res.status_code == 412

        # check non matching If-None-Match
        headers = [("Accept", "application/json"), ("If-None-Match", '"xyz", "def"')]
        for method in all_methods:
            res = method("/objects/1", headers=headers)
            check_normal_response(res, method)

        # check non matching If-None-Match with weak ETags
        headers = [
            ("Accept", "application/json"),
            ("If-None-Match", 'W/"xyz", W/"def"'),
        ]
        query_string_weak_etags = {"weak_etags": True}
        for method in all_methods:
            res = method(
                "/objects/1", headers=headers, query_string=query_string_weak_etags
            )
            check_normal_response(res, method)

        # check matching If-Match
        headers = [("Accept", "application/json"), ("If-Match", '"abc", "def"')]
        for method in all_methods:
            res = method("/objects/1", headers=headers)
            check_normal_response(res, method)

        # check matching If-Match with weak ETags
        headers = [("Accept", "application/json"), ("If-Match", 'W/"abc", W/"def"')]
        query_string_weak_etags = {"weak_etags": True}
        for method in all_methods:
            res = method(
                "/objects/1", headers=headers, query_string=query_string_weak_etags
            )
            check_normal_response(res, method)

        # check non matching If-Match
        headers = [("Accept", "application/json"), ("If-Match", '"xyz", "def"')]
        for method in all_methods:
            res = method("/objects/1", headers=headers)
            assert res.status_code == 412

        # check non matching If-Match with weak ETags
        headers = [("Accept", "application/json"), ("If-Match", 'W/"xyz", W/"def"')]
        query_string_weak_etags = {"weak_etags": True}
        for method in all_methods:
            res = method(
                "/objects/1", headers=headers, query_string=query_string_weak_etags
            )
            assert res.status_code == 412

        # check If-Modified-Since
        res = client.get("/objects/1")
        last_modified = res.headers["Last-Modified"]
        res = client.get("/objects/1", headers={"If-Modified-Since": last_modified})
        assert res.cache_control.no_cache
        assert res.status_code == 304



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Error module tests."""

from __future__ import absolute_import, print_function

import json

from invenio_rest import InvenioREST
from invenio_rest.errors import (
    FieldError,
    InvalidContentType,
    RESTException,
    RESTValidationError,
)


def test_errors(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route("/", methods=["GET"])
    def test_rest():
        raise RESTException(description="error description")

    @app.route("/contenttype", methods=["GET"])
    def test_content_type():
        raise InvalidContentType(allowed_content_types=["application/json"])

    @app.route("/validationerror", methods=["GET"])
    def test_validation_error():
        raise RESTValidationError(errors=[FieldError("myfield", "mymessage", code=10)])

    with app.test_client() as client:
        res = client.get("/")
        assert res.status_code == 200
        data = json.loads(res.get_data(as_text=True))
        assert data["status"] is None
        assert data["message"] == "error description"

        res = client.get("/contenttype")
        assert res.status_code == 415
        data = json.loads(res.get_data(as_text=True))
        assert data["status"] == 415
        assert "application/json" in data["message"]

        res = client.get("/validationerror")
        assert res.status_code == 400
        data = json.loads(res.get_data(as_text=True))
        print(data)
        assert data["status"] == 400
        assert data["message"] == "Validation error."
        assert data["errors"] == [dict(field="myfield", message="mymessage", code=10)]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C) 2024 Graz University of Technology.
# Copyright (C) 2025 University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import json

from flask import Blueprint, Flask, request

from invenio_rest.csrf import (
    REASON_BAD_REFERER,
    REASON_BAD_TOKEN,
    REASON_INSECURE_REFERER,
    REASON_MALFORMED_REFERER,
    REASON_NO_REFERER,
    CSRFProtectMiddleware,
    _get_new_csrf_token,
)


def test_csrf_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = CSRFProtectMiddleware(app)
    assert "invenio-csrf" in app.extensions

    app = Flask("testapp")
    ext = CSRFProtectMiddleware()
    assert "invenio-csrf" not in app.extensions
    ext.init_app(app)
    assert "invenio-csrf" in app.extensions


def test_csrf_disabled(csrf_app):
    """Test CSRF disabled."""
    with csrf_app.test_client() as client:
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200


def test_csrf_enabled(csrf_app, csrf):
    """Test CSRF enabled."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        # obtain a token
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.json["message"] == REASON_BAD_TOKEN
        assert res.status_code == 400

        # don't send the token
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.json["message"] == REASON_BAD_TOKEN
        assert res.status_code == 400

        # send the token
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        cookie = client.get_cookie(CSRF_COOKIE_NAME)

        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: cookie.value},
        )
        assert res.status_code == 200

        # The CSRF token should not have changed.
        new_cookie = client.get_cookie(CSRF_COOKIE_NAME)

        assert cookie.value == new_cookie.value


def test_csrf_before_csrf_protect(csrf_app, csrf):
    """Test before CSRF protect decorator."""
    assert csrf._before_protect_funcs == []

    @csrf.before_csrf_protect
    def before_protect():
        pass

    assert csrf._before_protect_funcs == [before_protect]

    csrf.before_csrf_protect(before_protect)

    assert csrf._before_protect_funcs == [before_protect, before_protect]


def test_csrf_exempt(csrf_app, csrf):
    """Test before CSRF protect decorator."""

    # Test `exempt` as a function passing the name of the view as string
    csrf.exempt("conftest.csrf_test")
    with csrf_app.test_client() as client:
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200


def test_csrf_exempt_dec(csrf_app, csrf):
    # Test `exempt` as a decorator on a view
    @csrf_app.route("/another-csrf-protect", methods=["POST"])
    @csrf.exempt
    def another_csrf_test():
        return "another test"

    with csrf_app.test_client() as client:
        res = client.post(
            "/another-csrf-protect",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200


def test_csrf_exempt_bp(csrf_app, csrf):
    # Test `exempt` as a decorator on a blueprint
    blueprint = Blueprint("test_csrf_bp", __name__, url_prefix="")

    @blueprint.route("/csrf-protect-bp", methods=["POST"])
    def csrf_bp():
        return "csrf bp test"

    @blueprint.route("/csrf-protect-bp-2", methods=["POST"])
    def csrf_bp_2():
        return "csrf bp test 2"

    csrf_app.register_blueprint(blueprint)

    csrf.exempt(blueprint)

    with csrf_app.test_client() as client:
        res = client.post(
            "/csrf-protect-bp",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200

        res = client.post(
            "/csrf-protect-bp-2",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200


def test_skip_csrf_check(csrf_app, csrf):
    """Test skipping CSRF check."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.json["message"] == REASON_BAD_TOKEN
        assert res.status_code == 400

        @csrf.before_csrf_protect
        def csrf_skip():
            request.skip_csrf_check = True

        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        assert res.status_code == 200


def test_csrf_not_signed_correctly(csrf_app, csrf):
    """Test CSRF malicious attempt with passing malicious cookie and header."""
    from invenio_base.jws import TimedJSONWebSignatureSerializer

    with csrf_app.test_client() as client:
        # try to pass our own signed cookie and header in an attempt to bypass
        # the CSRF check
        csrf_serializer = TimedJSONWebSignatureSerializer(
            "invalid_secret",
            salt="invalid_salt",
        )
        malicious_cookie = csrf_serializer.dumps({"user": "malicious"}, "my_secret")
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]
        client.set_cookie(
            CSRF_COOKIE_NAME, malicious_cookie.decode("utf-8"), domain="localhost"
        )

        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: malicious_cookie},
        )

        assert res.json["message"] == REASON_BAD_TOKEN
        assert res.status_code == 400


def test_csrf_token_rotation(csrf_app, csrf):
    """Test CSRF token rotation."""
    with csrf_app.test_client() as client:
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        # Token in grace period - succeeds but token gets rotated
        expired_token = _get_new_csrf_token(expires_in=-1)
        client.set_cookie(CSRF_COOKIE_NAME, expired_token, domain="localhost")
        old_cookie = client.get_cookie(CSRF_COOKIE_NAME)
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: expired_token},
        )
        assert res.status_code == 200
        # Token was rotated and new requests succeeds
        new_cookie = client.get_cookie(CSRF_COOKIE_NAME)
        assert new_cookie.value != old_cookie.value
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: new_cookie.value},
        )
        assert res.status_code == 200
        # Subsequent requests doesn't rotate CSRF token
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: new_cookie.value},
        )
        last_cookie = client.get_cookie(CSRF_COOKIE_NAME)
        assert new_cookie.value == last_cookie.value
        assert res.status_code == 200

        # Token outside grace period
        # - Hack to have a negative grace period to force the error.
        csrf_app.config["CSRF_TOKEN_GRACE_PERIOD"] = -10000
        expired_token = _get_new_csrf_token(expires_in=-60 * 60 * 24 * 14)
        client.set_cookie(CSRF_COOKIE_NAME, expired_token, domain="localhost")
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: expired_token},
        )
        assert res.status_code == 400


def test_csrf_no_referer(csrf_app, csrf):
    """Test CSRF no referrer in a secure request."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        # send the token to reach the CSRF check
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        cookie = client.get_cookie(CSRF_COOKIE_NAME)

        res = client.post(
            "/csrf-protected",
            base_url="https://localhost",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: cookie.value},
        )
        assert res.json["message"] == REASON_NO_REFERER
        assert res.status_code == 400


def test_csrf_malformed_referer(csrf_app, csrf):
    """Test CSRF malformed referrer in a secure request."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        # send the token to reach the CSRF check
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        cookie = client.get_cookie(CSRF_COOKIE_NAME)

        res = client.post(
            "/csrf-protected",
            base_url="https://localhost",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: cookie.value, "Referer": "bad-referer"},
        )
        assert res.json["message"] == REASON_MALFORMED_REFERER
        assert res.status_code == 400


def test_csrf_insecure_referer(csrf_app, csrf):
    """Test CSRF insecure referrer in a secure request."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        # send the token to reach the CSRF check
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        cookie = client.get_cookie(CSRF_COOKIE_NAME)

        res = client.post(
            "/csrf-protected",
            base_url="https://localhost",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={
                CSRF_HEADER_NAME: cookie.value,
                "Referer": "http://insecure-referer",
            },
        )
        assert res.json["message"] == REASON_INSECURE_REFERER
        assert res.status_code == 400


def test_csrf_bad_referer(csrf_app, csrf):
    """Test CSRF bad referrer in a secure request."""
    with csrf_app.test_client() as client:
        # First request to set the cookie
        res = client.post(
            "/csrf-protected",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
        )
        # send the token to reach the CSRF check
        CSRF_COOKIE_NAME = csrf_app.config["CSRF_COOKIE_NAME"]
        CSRF_HEADER_NAME = csrf_app.config["CSRF_HEADER"]

        cookie = client.get_cookie(CSRF_COOKIE_NAME)

        csrf_app.config["TRUSTED_HOSTS"] = ["allowed-referer"]
        not_allowed_referer = "https://not-allowed-referer"
        res = client.post(
            "/csrf-protected",
            base_url="https://localhost",
            data=json.dumps(dict(foo="bar")),
            content_type="application/json",
            headers={CSRF_HEADER_NAME: cookie.value, "Referer": not_allowed_referer},
        )
        assert res.json["message"] == REASON_BAD_REFERER % not_allowed_referer
        assert res.status_code == 400



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2025 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API module for Invenio."""

from __future__ import absolute_import, print_function

import warnings

from . import config
from .views import create_api_errorhandler


class InvenioREST(object):
    """Invenio-REST extension."""

    def __init__(self, app=None):
        """Extension initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization.

        Initialize the Rate-Limiter, CORS and error handlers.

        :param app: An instance of :class:`flask.Flask`.
        """
        self.init_config(app)

        # Enable CORS support if desired
        if app.config["REST_ENABLE_CORS"]:
            from flask_cors import CORS

            CORS(app)
            # CORS can be configured using CORS_* configuration variables.

        # Enable CSRF support if desired
        if app.config["REST_CSRF_ENABLED"]:
            from .csrf import csrf

            csrf.init_app(app)
        else:
            warnings.warn(
                "CSRF validation will be enabled by default in the version" " 1.3.x",
                category=FutureWarning,
                stacklevel=2,
            )

        app.errorhandler(400)(
            create_api_errorhandler(status=400, message="Bad Request")
        )
        app.errorhandler(401)(
            create_api_errorhandler(status=401, message="Unauthorized")
        )
        app.errorhandler(403)(create_api_errorhandler(status=403, message="Forbidden"))
        app.errorhandler(404)(create_api_errorhandler(status=404, message="Not Found"))
        app.errorhandler(405)(
            create_api_errorhandler(status=405, message="Method Not Allowed")
        )
        app.errorhandler(406)(
            create_api_errorhandler(status=406, message="Not Acceptable")
        )
        app.errorhandler(409)(create_api_errorhandler(status=409, message="Conflict"))
        app.errorhandler(410)(create_api_errorhandler(status=410, message="Gone"))
        app.errorhandler(412)(
            create_api_errorhandler(status=412, message="Precondition Failed")
        )
        app.errorhandler(415)(
            create_api_errorhandler(status=415, message="Unsupported media type")
        )
        app.errorhandler(422)(
            create_api_errorhandler(status=422, message="Unprocessable Entity")
        )
        app.errorhandler(429)(
            create_api_errorhandler(status=429, message="Rate limit exceeded")
        )
        app.errorhandler(500)(
            create_api_errorhandler(status=500, message="Internal Server Error")
        )
        app.errorhandler(501)(
            create_api_errorhandler(status=501, message="Not Implemented")
        )
        app.errorhandler(502)(
            create_api_errorhandler(status=502, message="Bad Gateway")
        )
        app.errorhandler(503)(
            create_api_errorhandler(status=503, message="Service Unavailable")
        )
        app.errorhandler(504)(
            create_api_errorhandler(status=504, message="Gateway Timeout")
        )

        app.extensions["invenio-rest"] = self

    def init_config(self, app):
        """Initialize configuration.

        .. note:: Change Flask-CORS and Flask-Limiter defaults.

        :param app: An instance of :class:`flask.Flask`.
        """
        config_apps = [
            "REST_",
            "CORS_",
        ]
        for k in dir(config):
            if any([k.startswith(prefix) for prefix in config_apps]):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2023-2024 Graz University of Technology.
# Copyright (C) 2025 University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CSRF Middleware.

The implementation is highly inspred from Django's initial implementation
about CSRF protection. For more information you can see here:
<https://github.com/django/django/blob/master/django/middleware/csrf.py>
"""


import secrets
import string
from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

from flask import Blueprint, abort, current_app, request
from invenio_base.jws import TimedJSONWebSignatureSerializer
from itsdangerous import BadSignature, SignatureExpired

REASON_NO_REFERER = "Referer checking failed - no Referer."
REASON_BAD_REFERER = "Referer checking failed - %s does not match any trusted origins."
REASON_NO_CSRF_COOKIE = "CSRF cookie not set."
REASON_BAD_TOKEN = "CSRF token missing or incorrect."
REASON_MALFORMED_REFERER = "Referer checking failed - Referer is malformed."
REASON_INSECURE_REFERER = (
    "Referer checking failed - Referer is insecure while host is secure."
)
REASON_TOKEN_EXPIRED = "CSRF token expired. Try again."


def _get_csrf_serializer(expires_in=None):
    """Note that this serializer is used to encode/decode the CSRF cookie.

    In case you change this implementation bear in mind that the token
    generated must be signed so as to avoid any client-side tampering.
    """
    expires_in = expires_in or current_app.config["CSRF_TOKEN_EXPIRES_IN"]

    return TimedJSONWebSignatureSerializer(
        current_app.config.get(
            "CSRF_SECRET", current_app.config.get("SECRET_KEY") or "CHANGE_ME"
        ),
        salt=current_app.config["CSRF_SECRET_SALT"],
        expires_in=expires_in,
    )


def _get_random_string(length, allowed_chars):
    return "".join(secrets.choice(allowed_chars) for i in range(length))


def _get_new_csrf_token(expires_in=None):
    csrf_serializer = _get_csrf_serializer(expires_in=expires_in)
    encoded_token = csrf_serializer.dumps(
        _get_random_string(
            current_app.config["CSRF_TOKEN_LENGTH"],
            current_app.config["CSRF_ALLOWED_CHARS"],
        )
    )
    return encoded_token.decode("utf-8")


def _get_csrf_token():
    try:
        csrf_cookie = request.cookies[current_app.config["CSRF_COOKIE_NAME"]]
    except KeyError:
        return None
    return _decode_csrf(csrf_cookie)


def _decode_csrf(data):
    csrf_serializer = _get_csrf_serializer()
    try:
        return csrf_serializer.loads(data)
    except SignatureExpired as e:
        grace_period = timedelta(seconds=current_app.config["CSRF_TOKEN_GRACE_PERIOD"])
        # Because we support ItsDangerous 1.X and 2.X we need to compute the
        # appropriately-timezone-aware-or-naive datetime. See
        # https://github.com/pallets/itsdangerous/blob/1.0.x/src/itsdangerous/jws.py#L212  # noqa
        # https://github.com/pallets/itsdangerous/blob/2.0.x/src/itsdangerous/jws.py#L244  # noqa
        now = (
            datetime.now(tz=timezone.utc) if e.date_signed.tzinfo else datetime.utcnow()
        )  # noqa
        if e.date_signed < now - grace_period:
            # Grace period for token rotation exceeded.
            _abort400(REASON_TOKEN_EXPIRED)
        else:
            # Accept expired token, but rotate it to a new one.
            reset_token()
            return e.payload
    except BadSignature:
        _abort400(REASON_BAD_TOKEN)


def _set_token(response):
    response.set_cookie(
        current_app.config["CSRF_COOKIE_NAME"],
        _get_new_csrf_token(),
        max_age=current_app.config.get(
            # 1 week for cookie (but we rotate the token every day)
            "CSRF_COOKIE_MAX_AGE",
            60 * 60 * 24 * 7,
        ),
        domain=current_app.config.get(
            "CSRF_COOKIE_DOMAIN",
            current_app.session_interface.get_cookie_domain(current_app),
        ),
        path=current_app.session_interface.get_cookie_path(current_app),
        secure=current_app.config.get("SESSION_COOKIE_SECURE", True),
        httponly=False,
        samesite=current_app.config["CSRF_COOKIE_SAMESITE"],
    )


def _get_submitted_csrf_token():
    header_name = current_app.config["CSRF_HEADER"]
    csrf_token = request.headers.get(header_name)
    if csrf_token:
        return csrf_token
    return None


def _is_referer_secure(referer):
    return (
        "https" in referer.scheme or not current_app.config["CSRF_FORCE_SECURE_REFERER"]
    )


def _abort400(reason):
    abort(400, reason)


def csrf_validate():
    """Check CSRF cookie against request headers."""
    # If the cookie is not set, we don't need to check anything.
    if not request.cookies:
        return

    csrf_token = _get_csrf_token()
    if csrf_token is None:
        return _abort400(REASON_NO_CSRF_COOKIE)

    request_csrf_token = _get_submitted_csrf_token()
    if not request_csrf_token:
        _abort400(REASON_BAD_TOKEN)

    if request.is_secure:
        referer = request.referrer
        if referer is None:
            return _abort400(REASON_NO_REFERER)

        referer = urlparse(referer)
        # Make sure we have a valid URL for Referer.
        if "" in (referer.scheme, referer.netloc):
            return _abort400(REASON_MALFORMED_REFERER)

        # Ensure that our Referer is also secure.
        if not _is_referer_secure(referer):
            return _abort400(REASON_INSECURE_REFERER)

        is_hostname_allowed = referer.hostname in current_app.config.get(
            "TRUSTED_HOSTS"
        )
        if not is_hostname_allowed:
            reason = REASON_BAD_REFERER % referer.geturl()
            return _abort400(reason)

    decoded_request_csrf_token = _decode_csrf(request_csrf_token)

    if csrf_token != decoded_request_csrf_token:
        return _abort400(REASON_BAD_TOKEN)


def reset_token():
    """Change the CSRF token in use for a request."""
    request.csrf_cookie_needs_reset = True


class CSRFTokenMiddleware:
    """CSRF Token Middleware."""

    def __init__(self, app=None):
        """Middleware initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize middleware extension.

        :param app: An instance of :class:`flask.Flask`.
        """
        app.config.setdefault("CSRF_COOKIE_NAME", "csrftoken")
        app.config.setdefault("CSRF_HEADER", "X-CSRFToken")
        app.config.setdefault("CSRF_METHODS", ["POST", "PUT", "PATCH", "DELETE"])
        app.config.setdefault("CSRF_TOKEN_LENGTH", 32)
        app.config.setdefault(
            "CSRF_ALLOWED_CHARS", string.ascii_letters + string.digits
        )
        app.config.setdefault("CSRF_SECRET_SALT", "invenio-csrf-token")
        app.config.setdefault("CSRF_FORCE_SECURE_REFERER", True)
        app.config.setdefault(
            "CSRF_COOKIE_SAMESITE", app.config.get("SESSION_COOKIE_SAMESITE") or "Lax"
        )
        # The token last for 24 hours, but the cookie for 7 days. This allows
        # us to implement transparent token rotation during those 7 days. Note,
        # that the token is automatically rotated on login, thus you can also
        # change PERMANENT_SESSION_LIFETIME
        app.config.setdefault("CSRF_TOKEN_EXPIRES_IN", 60 * 60 * 24)
        # We allow usage of an expired CSRF token during this period. This way
        # we can rotate the CSRF token without the user getting an CSRF error.
        # Align with CSRF_COOKIE_MAX_AGE
        app.config.setdefault("CSRF_TOKEN_GRACE_PERIOD", 60 * 60 * 24 * 7)

        @app.after_request
        def csrf_send(response):
            is_method_vulnerable = request.method in app.config["CSRF_METHODS"]
            cookie_needs_reset = getattr(request, "csrf_cookie_needs_reset", False)
            cookie_is_missing = (
                current_app.config["CSRF_COOKIE_NAME"] not in request.cookies
            )
            if (is_method_vulnerable and cookie_is_missing) or cookie_needs_reset:
                _set_token(response)
            return response

        app.extensions["invenio-csrf"] = self


class CSRFProtectMiddleware(CSRFTokenMiddleware):
    """CSRF Middleware."""

    def __init__(self, app=None):
        """Middleware initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        self._exempt_views = set()
        self._exempt_blueprints = set()

        self._before_protect_funcs = []
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize middleware extension.

        :param app: An instance of :class:`flask.Flask`.
        """
        super(CSRFProtectMiddleware, self).init_app(app)

        @app.before_request
        def csrf_protect():
            """CSRF protect method."""
            for func in self._before_protect_funcs:
                func()

            is_method_vulnerable = request.method in app.config["CSRF_METHODS"]
            if not is_method_vulnerable:
                return

            if request.blueprint in self._exempt_blueprints:
                return

            if hasattr(request, "skip_csrf_check"):
                return

            view = app.view_functions.get(request.endpoint)
            if view:
                dest = "{0}.{1}".format(view.__module__, view.__name__)
                if dest in self._exempt_views:
                    return

            return csrf_validate()

    def before_csrf_protect(self, f):
        """Register functions to be invoked before checking csrf.

        The function accepts nothing as parameters.
        """
        self._before_protect_funcs.append(f)
        return f

    def exempt(self, view):
        """Mark a view or blueprint to be excluded from CSRF protection."""
        if isinstance(view, Blueprint):
            self._exempt_blueprints.add(view.name)
            return view

        if isinstance(view, str):
            view_location = view
        else:
            view_location = ".".join((view.__module__, view.__name__))

        self._exempt_views.add(view_location)
        return view


csrf = CSRFProtectMiddleware()
